@@ -3,15 +3,10 @@
 Hierarchical Firewalls (HF) are represented in a SecurityPolicy GCP resouce.
 """
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
 import copy
 import re
 
-from typing import Dict, Text, Any
+from typing import Dict, Any
 
 from absl import logging
 from capirca.lib import gcp
@@ -26,6 +21,28 @@ class DifferentPolicyNameError(gcp.Error):
   """Raised when headers in the same policy have a different policy name."""
 
 
+class ApiVersionSyntaxMap:
+  """Defines the syntax changes between different API versions.
+
+   http://cloud/compute/docs/reference/rest/v1/firewallPolicies/addRule
+   http://cloud/compute/docs/reference/rest/beta/organizationSecurityPolicies/addRule
+  """
+  SYNTAX_MAP = {
+      'beta': {
+          'display_name': 'displayName',
+          'dest_ip_range': 'destIpRanges',
+          'src_ip_range': 'srcIpRanges',
+          'layer_4_config': 'layer4Configs'
+      },
+      'ga': {
+          'display_name': 'shortName',
+          'dest_ip_range': 'destIpRanges',
+          'src_ip_range': 'srcIpRanges',
+          'layer_4_config': 'layer4Configs'
+      }
+  }
+
+
 class Term(gcp.Term):
   """Used to create an individual term."""
 
@@ -33,8 +50,6 @@ class Term(gcp.Term):
 
   _MAX_TERM_COMMENT_LENGTH = 64
 
-  _PROTO_NAMES = ['tcp', 'udp', 'icmp', 'icmpv6', 'esp', 'ah', 'sctp']
-
   _TARGET_RESOURCE_FORMAT = 'https://www.googleapis.com/compute/v1/projects/{}/global/networks/{}'
 
   _TERM_ADDRESS_LIMIT = 256
@@ -43,40 +58,27 @@ class Term(gcp.Term):
 
   _TERM_DESTINATION_PORTS_LIMIT = 256
 
-  def __init__(self, term, address_family='inet'):
-    super(Term, self).__init__(term)
+  def __init__(self,
+               term,
+               address_family='inet',
+               policy_inet_version='inet',
+               api_version='beta'):
+    super().__init__(term)
     self.address_family = address_family
     self.term = term
     self.skip = False
     self._ValidateTerm()
+    self.api_version = api_version
 
-    # Don't render icmp protocol terms under inet6.
-    if self.address_family == 'inet6':
-      if ['icmp'] == self.term.protocol:
-        # Skip term if its only protocol is icmp to prevent an empty list,
-        # which is equivalent to any protocol.
-        self.skip = True
-      elif 'icmp' in self.term.protocol:
-        self.term.protocol.remove('icmp')
-
-    # Don't render icmpv6 protocol terms under inet.
-    if self.address_family == 'inet':
-      # Skip term if its only protocol is icmpv6 to prevent an empty list,
-      # which is equivalent to any protocol.
-      if ['icmpv6'] == self.term.protocol:
-        self.skip = True
-      elif 'icmpv6' in self.term.protocol:
-        self.term.protocol.remove('icmpv6')
+    # This is to handle mixed, where the policy_inet_version is mixed,
+    # but the term inet version is either inet/inet6.
+    # This is only useful for term name and priority.
+    self.policy_inet_version = policy_inet_version
 
   def _ValidateTerm(self):
     if self.term.destination_tag or self.term.source_tag:
       raise gcp.TermError('Hierarchical Firewall does not support tags')
 
-    if self.term.protocol:
-      for protocol in self.term.protocol:
-        if protocol not in self._PROTO_NAMES:
-          raise gcp.TermError('Protocol %s is not supported' % protocol)
-
     if len(self.term.target_resources) > self._TERM_TARGET_RESOURCES_LIMIT:
       raise gcp.TermError(
           'Term: %s  target_resources field contains %s resources. It should not contain more than "%s".'
@@ -106,6 +108,15 @@ class Term(gcp.Term):
           % (self.term.name, str(len(
               self.term.destination_port)), self._TERM_DESTINATION_PORTS_LIMIT))
 
+    # Since policy_inet_version is used to handle 'mixed'.
+    # We should error out if the individual term's inet version (address_family)
+    # is anything other than inet/inet6, since this should never happen
+    # naturally. Something has gone horribly wrong if you encounter this error.
+    if self.address_family == 'mixed':
+      raise gcp.TermError(
+          'Hierarchical firewall rule has incorrect inet_version for rule: %s' %
+          self.term.name)
+
   def ConvertToDict(self, priority_index):
     """Converts term to dict representation of SecurityPolicy.Rule JSON format.
 
@@ -124,12 +135,25 @@ class Term(gcp.Term):
 
     rules = []
 
+    # Identify if this is inet6 processing for a term under a mixed policy.
+    mixed_policy_inet6_term = False
+    if self.policy_inet_version == 'mixed' and self.address_family == 'inet6':
+      mixed_policy_inet6_term = True
+
     term_dict = {
         'action': self.ACTION_MAP.get(self.term.action[0], self.term.action[0]),
         'direction': self.term.direction,
         'priority': priority_index
     }
 
+    # Get the correct syntax for API versions.
+    src_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[
+        self.api_version]['src_ip_range']
+    dest_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[
+        self.api_version]['dest_ip_range']
+    layer_4_config = ApiVersionSyntaxMap.SYNTAX_MAP[
+        self.api_version]['layer_4_config']
+
     target_resources = []
     for proj, vpc in self.term.target_resources:
       target_resources.append(self._TARGET_RESOURCE_FORMAT.format(proj, vpc))
@@ -140,17 +164,51 @@ class Term(gcp.Term):
     term_dict['enableLogging'] = self._GetLoggingSetting()
 
     # This combo provides ability to identify the rule.
-    raw_descirption = self.term.name + ': ' + ' '.join(self.term.comment)
-    term_dict['description'] = gcp.TruncateString(raw_descirption,
+    term_name = self.term.name
+    if mixed_policy_inet6_term:
+      term_name = gcp.GetIpv6TermName(term_name)
+    raw_description = term_name + ': ' + ' '.join(self.term.comment)
+    term_dict['description'] = gcp.TruncateString(raw_description,
                                                   self._MAX_TERM_COMMENT_LENGTH)
 
+    filtered_protocols = []
+    for proto in self.term.protocol:
+      # ICMP filtering by inet_version
+      # Since each term has inet_version, 'mixed' is correctly processed here.
+      if proto == 'icmp' and self.address_family == 'inet6':
+        logging.warning(
+            'WARNING: Term %s is being rendered for inet6, ICMP '
+            'protocol will not be rendered.', self.term.name)
+        continue
+      if proto == 'icmpv6' and self.address_family == 'inet':
+        logging.warning(
+            'WARNING: Term %s is being rendered for inet, ICMPv6 '
+            'protocol will not be rendered.', self.term.name)
+        continue
+      if proto == 'igmp' and self.address_family == 'inet6':
+        logging.warning(
+            'WARNING: Term %s is being rendered for inet6, IGMP '
+            'protocol will not be rendered.', self.term.name)
+        continue
+      filtered_protocols.append(proto)
+    # If there is no protocol left after ICMP/IGMP filtering, drop this term.
+    # But only do this for terms that originally had protocols.
+    # Otherwise you end up dropping the default-deny.
+    if self.term.protocol and not filtered_protocols:
+      return {}
+
     protocols_and_ports = []
     if not self.term.protocol:
       # Empty protocol list means any protocol, but any protocol in HF is
       # represented as "all"
       protocols_and_ports = [{'ipProtocol': 'all'}]
     else:
-      for proto in self.term.protocol:
+      for proto in filtered_protocols:
+        # If the protocol name is not supported, use the protocol number.
+        if proto not in self._ALLOW_PROTO_NAME:
+          proto = self.PROTO_MAP[proto]
+          logging.info('INFO: Term %s is being rendered using protocol number',
+                       self.term.name)
         proto_ports = {'ipProtocol': proto}
         if self.term.destination_port:
           ports = self._GetPorts()
@@ -158,11 +216,10 @@ class Term(gcp.Term):
             proto_ports['ports'] = ports
         protocols_and_ports.append(proto_ports)
 
-    term_dict['match'] = {
-        'config': {
-            'layer4Configs': protocols_and_ports
-        }
-    }
+    if self.api_version == 'ga':
+      term_dict['match'] = {layer_4_config: protocols_and_ports}
+    else:
+      term_dict['match'] = {'config': {layer_4_config: protocols_and_ports}}
 
     # match needs a field called versionedExpr with value FIREWALL
     # See documentation:
@@ -176,35 +233,69 @@ class Term(gcp.Term):
       any_ip = [nacaddr.IPv6('::/0')]
 
     if self.term.direction == 'EGRESS':
-      daddrs = self.term.GetAddressOfVersion('destination_address',
-                                             ip_version)
+      daddrs = self.term.GetAddressOfVersion('destination_address', ip_version)
+
+      # If the address got filtered out and is empty due to address family, we
+      # don't render the term. At this point of term processing, the direction
+      # has already been validated, so we can just log and return empty rule.
+      if self.term.destination_address and not daddrs:
+        logging.warning(
+            'WARNING: Term %s is not being rendered for %s, '
+            'because there are no addresses of that family.', self.term.name,
+            self.address_family)
+        return []
+      # This should only happen if there were no addresses set originally.
       if not daddrs:
         daddrs = any_ip
 
       destination_address_chunks = [
-          daddrs[x:x+self._TERM_ADDRESS_LIMIT] for x in
-          range(0, len(daddrs), self._TERM_ADDRESS_LIMIT)]
+          daddrs[x:x + self._TERM_ADDRESS_LIMIT]
+          for x in range(0, len(daddrs), self._TERM_ADDRESS_LIMIT)
+      ]
 
       for daddr_chunk in destination_address_chunks:
         rule = copy.deepcopy(term_dict)
-        rule['match']['config']['destIpRanges'] = [daddr.with_prefixlen for
-                                                   daddr in daddr_chunk]
+        if self.api_version == 'ga':
+          rule['match'][dest_ip_range] = [
+              daddr.with_prefixlen for daddr in daddr_chunk
+          ]
+        else:
+          rule['match']['config'][dest_ip_range] = [
+              daddr.with_prefixlen for daddr in daddr_chunk
+          ]
         rule['priority'] = priority_index
         rules.append(rule)
         priority_index += 1
     else:
       saddrs = self.term.GetAddressOfVersion('source_address', ip_version)
 
+      # If the address got filtered out and is empty due to address family, we
+      # don't render the term. At this point of term processing, the direction
+      # has already been validated, so we can just log and return empty rule.
+      if self.term.source_address and not saddrs:
+        logging.warning(
+            'WARNING: Term %s is not being rendered for %s, '
+            'because there are no addresses of that family.', self.term.name,
+            self.address_family)
+        return []
+      # This should only happen if there were no addresses set originally.
       if not saddrs:
         saddrs = any_ip
 
       source_address_chunks = [
-          saddrs[x:x+self._TERM_ADDRESS_LIMIT] for x in
-          range(0, len(saddrs), self._TERM_ADDRESS_LIMIT)]
+          saddrs[x:x + self._TERM_ADDRESS_LIMIT]
+          for x in range(0, len(saddrs), self._TERM_ADDRESS_LIMIT)
+      ]
       for saddr_chunk in source_address_chunks:
         rule = copy.deepcopy(term_dict)
-        rule['match']['config']['srcIpRanges'] = [saddr.with_prefixlen for
-                                                  saddr in saddr_chunk]
+        if self.api_version == 'ga':
+          rule['match'][src_ip_range] = [
+              saddr.with_prefixlen for saddr in saddr_chunk
+          ]
+        else:
+          rule['match']['config'][src_ip_range] = [
+              saddr.with_prefixlen for saddr in saddr_chunk
+          ]
         rule['priority'] = priority_index
         rules.append(rule)
         priority_index += 1
@@ -219,9 +310,14 @@ class HierarchicalFirewall(gcp.GCP):
   """A GCP Hierarchical Firewall policy."""
 
   SUFFIX = '.gcphf'
-  _ANY = [nacaddr.IP('0.0.0.0/0')]
+  _ANY_IP = {
+      'inet': nacaddr.IP('0.0.0.0/0'),
+      'inet6': nacaddr.IP('::/0'),
+  }
   _PLATFORM = 'gcp_hf'
-  _SUPPORTED_AF = frozenset(['inet'])
+  _SUPPORTED_AF = frozenset(['inet', 'inet6', 'mixed'])
+  # Beta is the default API version. GA supports IPv6 (inet6/mixed).
+  _SUPPORTED_API_VERSION = frozenset(['beta', 'ga'])
   _DEFAULT_MAXIMUM_COST = 100
 
   def _BuildTokens(self):
@@ -230,7 +326,7 @@ class HierarchicalFirewall(gcp.GCP):
     Returns:
       Tuple containing both supported tokens and sub tokens.
     """
-    supported_tokens, _ = super(HierarchicalFirewall, self)._BuildTokens()
+    supported_tokens, _ = super()._BuildTokens()
 
     supported_tokens |= {
         'destination_tag', 'expiration', 'source_tag', 'translated',
@@ -238,8 +334,8 @@ class HierarchicalFirewall(gcp.GCP):
     }
 
     supported_tokens -= {
-        'destination_address_exclude', 'expiration', 'icmp_type', 'platform',
-        'platform_exclude', 'source_address_exclude', 'verbatim'
+        'destination_address_exclude', 'expiration', 'icmp_type',
+        'source_address_exclude', 'verbatim'
     }
 
     supported_sub_tokens = {'action': {'accept', 'deny', 'next'}}
@@ -295,6 +391,14 @@ class HierarchicalFirewall(gcp.GCP):
           address_family = i
           filter_options.remove(i)
 
+      # Get the compute API version if set.
+      api_version = 'beta'
+      for i in self._SUPPORTED_API_VERSION:
+        if i in filter_options:
+          api_version = i
+          filter_options.remove(i)
+          break
+
       # Find the default maximum cost of a policy, an integer, if specified.
       max_cost = self._DEFAULT_MAXIMUM_COST
       for opt in filter_options:
@@ -309,6 +413,12 @@ class HierarchicalFirewall(gcp.GCP):
         raise gcp.HeaderError(
             'Default maximum cost cannot be higher than 65536')
 
+      if api_version == 'beta' and address_family != 'inet':
+        raise gcp.HeaderError(
+            'Beta compute API does not support IPv6, only inet is supported')
+
+      display_name = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['display_name']
+
       # Get policy name and validate it to meet displayName requirements.
       policy_name = header.FilterName(self._PLATFORM)
       if not policy_name:
@@ -325,12 +435,12 @@ class HierarchicalFirewall(gcp.GCP):
             'a lowercase letter, and all following characters must be a dash, '
             'lowercase letter, or digit, except the last character, which '
             'cannot be a dash.' % (policy_name))
-      if 'displayName' in policy and policy['displayName'] != policy_name:
+      if display_name in policy and policy[display_name] != policy_name:
         raise DifferentPolicyNameError(
             'Policy names that are from the same policy are expected to be '
             'equal, but %s is different to %s' %
-            (policy['displayName'], policy_name))
-      policy['displayName'] = policy_name
+            (policy[display_name], policy_name))
+      policy[display_name] = policy_name
 
       # If there are remaining options, they are unknown/unsupported options.
       if filter_options:
@@ -338,31 +448,63 @@ class HierarchicalFirewall(gcp.GCP):
             'Unsupported or unknown filter options %s in policy %s ' %
             (str(filter_options), policy_name))
 
+      # Handle mixed for each indvidual term as inet and inet6.
+      # inet/inet6 are treated the same.
+      term_address_families = []
+      if address_family == 'mixed':
+        term_address_families = ['inet', 'inet6']
+      else:
+        term_address_families = [address_family]
+
       for term in terms:
         if term.stateless_reply:
           continue
 
         if gcp.IsDefaultDeny(term):
           if direction == 'EGRESS':
-            term.destination_address = self._ANY
+            if address_family != 'mixed':
+              # Default deny also gets processed as part of terms processing.
+              # The name and priority get updated there.
+              term.destination_address = [self._ANY_IP[address_family]]
+            else:
+              term.destination_address = [
+                  self._ANY_IP['inet'], self._ANY_IP['inet6']
+              ]
           else:
-            term.source_address = self._ANY
+            if address_family != 'mixed':
+              term.source_address = [self._ANY_IP[address_family]]
+            else:
+              term.source_address = [
+                  self._ANY_IP['inet'], self._ANY_IP['inet6']
+              ]
         term.name = self.FixTermLength(term.name)
         term.direction = direction
 
-        rules = Term(term, address_family=address_family).ConvertToDict(
-            priority_index=counter)
-        if not rules:
-          continue
-        for dict_term in rules:
-          total_cost += GetRuleTupleCount(dict_term)
-          if total_cost > max_cost:
-            raise ExceededCostError('Policy cost (%d) for %s reached the '
-                                    'maximum (%d)' % (
-                                        total_cost, policy['displayName'],
-                                        max_cost))
-          policy['rules'].append(dict_term)
-        counter += len(rules)
+        # Only generate the term if it's for the appropriate platform
+        if term.platform:
+          if self._PLATFORM not in term.platform:
+            continue
+        if term.platform_exclude:
+          if self._PLATFORM in term.platform_exclude:
+            continue
+
+        for term_af in term_address_families:
+          rules = Term(
+              term,
+              address_family=term_af,
+              policy_inet_version=address_family,
+              api_version=api_version).ConvertToDict(priority_index=counter)
+          if not rules:
+            continue
+          for dict_term in rules:
+            total_cost += GetRuleTupleCount(dict_term, api_version)
+            if total_cost > max_cost:
+              raise ExceededCostError(
+                  'Policy cost (%d) for %s reached the '
+                  'maximum (%d)' %
+                  (total_cost, policy[display_name], max_cost))
+            policy['rules'].append(dict_term)
+          counter += len(rules)
 
     self.policies.append(policy)
 
@@ -372,10 +514,10 @@ class HierarchicalFirewall(gcp.GCP):
 
     if total_cost > 0:
       logging.info('Policy %s quota cost: %d',
-                   policy['displayName'], total_cost)
+                   policy[display_name], total_cost)
 
 
-def GetRuleTupleCount(dict_term: Dict[Text, Any]):
+def GetRuleTupleCount(dict_term: Dict[str, Any], api_version):
   """Calculate the tuple count of a rule in its dictionary form.
 
   Quota is charged based on how complex the rules are rather than simply
@@ -389,17 +531,25 @@ def GetRuleTupleCount(dict_term: Dict[Text, Any]):
 
   Args:
     dict_term: A dict object.
+    api_version: A string indicating the api version.
 
   Returns:
     int: The tuple count of the rule.
   """
-  config = dict_term.get('match', {}).get('config', {})
-  addresses_count = len(
-      config.get('destIpRanges', []) + config.get('srcIpRanges', []))
   layer4_count = 0
+  layer_4_config = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['layer_4_config']
+  dest_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['dest_ip_range']
+  src_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['src_ip_range']
   targets_count = len(dict_term.get('targetResources', []))
+  if api_version == 'ga':
+    config = dict_term.get('match', {})
+  else:
+    config = dict_term.get('match', {}).get('config', {})
 
-  for l4config in config.get('layer4Configs', []):
+  addresses_count = len(
+      config.get(dest_ip_range, []) + config.get(src_ip_range, []))
+
+  for l4config in config.get(layer_4_config, []):
     for _ in l4config.get('ports', []):
       layer4_count += 1
     if l4config.get('ipProtocol'):
