@@ -15,18 +15,12 @@
 
 """Juniper JCL generator."""
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
 import datetime
 from absl import logging
 from capirca.lib import aclgenerator
 from capirca.lib import nacaddr
 from capirca.lib import summarizer
 import six
-from six.moves import range
 
 
 # generic error class
@@ -70,7 +64,7 @@ class JuniperFragmentInV6Error(Error):
   pass
 
 
-class Config(object):
+class Config:
   """Config allows a configuration to be assembled easily.
 
   Configurations are automatically indented following Juniper's style.
@@ -169,23 +163,25 @@ class Term(aclgenerator.Term):
                            'daddr': 'ip-destination-address',
                            'protocol': 'ip-protocol',
                            'protocol-except': 'ip-protocol-except',
-                           'tcp-est': 'tcp-flags "(ack|rst)"'}}
+                           'tcp-est': 'tcp-flags "(ack|rst)"'}
+              }
 
   def __init__(self, term, term_type, enable_dsmo, noverbose):
-    super(Term, self).__init__(term)
+    super().__init__(term)
     self.term = term
     self.term_type = term_type
     self.enable_dsmo = enable_dsmo
     self.noverbose = noverbose
 
-    if term_type not in self._TERM_TYPE:
-      raise ValueError('Unknown Filter Type: %s' % term_type)
-    if 'hopopt' in self.term.protocol:
-      loc = self.term.protocol.index('hopopt')
-      self.term.protocol[loc] = 'hop-by-hop'
+    if self._PLATFORM != 'msmpc':
+      if term_type not in self._TERM_TYPE:
+        raise ValueError('Unknown Filter Type: %s' % term_type)
+      if 'hopopt' in self.term.protocol:
+        loc = self.term.protocol.index('hopopt')
+        self.term.protocol[loc] = 'hop-by-hop'
 
-    # some options need to modify the actions
-    self.extra_actions = []
+      # some options need to modify the actions
+      self.extra_actions = []
 
   # TODO(pmoody): get rid of all of the default string concatenation here.
   #  eg, indent(8) + 'foo;' -> '%s%s;' % (indent(8), 'foo'). pyglint likes this
@@ -283,7 +279,13 @@ class Term(aclgenerator.Term):
       term_prefix = ''
 
     # term name
-    config.Append('%s term %s {' %(term_prefix, self.term.name))
+    config.Append('%s term %s {' % (term_prefix, self.term.name))
+
+    # The "filter" keyword is not compatible with from or then
+    if self.term.filter_term:
+      config.Append('filter %s;' % self.term.filter_term)
+      config.Append('}')  # end term accept-foo-to-bar { ... }
+      return str(config)
 
     # a default action term doesn't have any from { clause
     has_match_criteria = (self.term.address or
@@ -435,7 +437,8 @@ class Term(aclgenerator.Term):
           config.Append(epfx + ' except;')
         config.Append('}')
 
-      if self.term.ttl:
+      # Only generate ttl if inet, inet6 uses hop-limit instead.
+      if self.term.ttl and self.term_type == 'inet':
         config.Append('ttl %s;' % self.term.ttl)
 
       # protocol
@@ -448,8 +451,7 @@ class Term(aclgenerator.Term):
       # protocol
       if self.term.protocol_except:
         # same as above
-        config.Append(family_keywords['protocol-except'] + ' '
-                      + self._Group(self.term.protocol_except))
+        config.Append(family_keywords['protocol-except'] + ' ' + self._Group(self.term.protocol_except))
 
       # port
       if self.term.port:
@@ -829,7 +831,7 @@ class Term(aclgenerator.Term):
         string: either the lower()'ed string or the ports, hyphenated
                 if they're a range, or by itself if it's not.
       """
-      if isinstance(el, str) or isinstance(el, six.text_type):
+      if isinstance(el, str) or isinstance(el, str):
         if lc:
           return el
         else:
@@ -861,7 +863,7 @@ class Juniper(aclgenerator.ACLGenerator):
 
   _PLATFORM = 'juniper'
   _DEFAULT_PROTOCOL = 'ip'
-  _SUPPORTED_AF = set(('inet', 'inet6', 'bridge'))
+  _SUPPORTED_AF = frozenset(('inet', 'inet6', 'bridge', 'mixed'))
   _TERM = Term
   SUFFIX = '.jcl'
 
@@ -871,9 +873,10 @@ class Juniper(aclgenerator.ACLGenerator):
     Returns:
       tuple containing both supported tokens and sub tokens
     """
-    supported_tokens, supported_sub_tokens = super(Juniper, self)._BuildTokens()
+    supported_tokens, supported_sub_tokens = super()._BuildTokens()
 
     supported_tokens |= {'address',
+                         'restrict_address_family',
                          'counter',
                          'destination_prefix',
                          'destination_prefix_except',
@@ -882,6 +885,7 @@ class Juniper(aclgenerator.ACLGenerator):
                          'dscp_set',
                          'encapsulate',
                          'ether_type',
+                         'filter_term',
                          'flexible_match_range',
                          'forwarding_class',
                          'forwarding_class_except',
@@ -915,7 +919,7 @@ class Juniper(aclgenerator.ACLGenerator):
             'tcp-established',
             'tcp-initial',
             'inactive'}
-        })
+         })
     return supported_tokens, supported_sub_tokens
 
   def _TranslatePolicy(self, pol, exp_info):
@@ -946,42 +950,61 @@ class Juniper(aclgenerator.ACLGenerator):
       if len(filter_options) > 1:
         filter_type = filter_options[1]
 
-      term_names = set()
-      new_terms = []
-      for term in terms:
+      if filter_type == 'mixed':
+        filter_types_to_process = ['inet', 'inet6']
+      else:
+        filter_types_to_process = [filter_type]
 
-        # if inactive is set, deactivate the term and remove the option.
-        if 'inactive' in term.option:
-          term.inactive = True
-          term.option.remove('inactive')
+      for filter_type in filter_types_to_process:
 
-        term.name = self.FixTermLength(term.name)
+        filter_name_suffix = ''
+        # If mixed filter_type, will append 4 or 6 to the filter name
+        if len(filter_types_to_process) > 1:
+          if filter_type == 'inet':
+            filter_name_suffix = '4'
+          if filter_type == 'inet6':
+            filter_name_suffix = '6'
 
-        if term.name in term_names:
-          raise JuniperDuplicateTermError('You have multiple terms named: %s' %
-                                          term.name)
-        term_names.add(term.name)
+        term_names = set()
+        new_terms = []
+        for term in terms:
 
-        term = self.FixHighPorts(term, af=filter_type)
-        if not term:
-          continue
-
-        if term.expiration:
-          if term.expiration <= exp_info_date:
-            logging.info('INFO: Term %s in policy %s expires '
-                         'in less than two weeks.', term.name, filter_name)
-          if term.expiration <= current_date:
-            logging.warning('WARNING: Term %s in policy %s is expired and '
-                            'will not be rendered.', term.name, filter_name)
+          # Ignore if the term is for a different AF
+          if term.restrict_address_family and term.restrict_address_family != filter_type:
             continue
-        if 'is-fragment' in term.option and filter_type == 'inet6':
-          raise JuniperFragmentInV6Error('The term %s uses "is-fragment" but '
-                                         'is a v6 policy.' % term.name)
 
-        new_terms.append(self._TERM(term, filter_type, enable_dsmo, noverbose))
+          # if inactive is set, deactivate the term and remove the option.
+          if 'inactive' in term.option:
+            term.inactive = True
+            term.option.remove('inactive')
 
-      self.juniper_policies.append((header, filter_name, filter_type,
-                                    interface_specific, new_terms))
+          term.name = self.FixTermLength(term.name)
+
+          if term.name in term_names:
+            raise JuniperDuplicateTermError('You have multiple terms named: %s' %
+                                            term.name)
+          term_names.add(term.name)
+
+          term = self.FixHighPorts(term, af=filter_type)
+          if not term:
+            continue
+
+          if term.expiration:
+            if term.expiration <= exp_info_date:
+              logging.info('INFO: Term %s in policy %s expires '
+                          'in less than two weeks.', term.name, filter_name)
+            if term.expiration <= current_date:
+              logging.warning('WARNING: Term %s in policy %s is expired and '
+                              'will not be rendered.', term.name, filter_name)
+              continue
+          if 'is-fragment' in term.option and filter_type == 'inet6':
+            raise JuniperFragmentInV6Error('The term %s uses "is-fragment" but '
+                                          'is a v6 policy.' % term.name)
+
+          new_terms.append(self._TERM(term, filter_type, enable_dsmo, noverbose))
+
+        self.juniper_policies.append((header, filter_name + filter_name_suffix, filter_type,
+                                      interface_specific, new_terms))
 
   def __str__(self):
     config = Config()
