@@ -1,4 +1,3 @@
-# Lint as: python2, python3
 # Copyright 2017 Google Inc. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -15,6 +14,7 @@
 """Palo Alto Firewall generator."""
 
 import collections
+import copy
 import datetime
 import logging
 import re
@@ -65,128 +65,108 @@ class PaloAltoFWBadIcmpTypeError(Error):
   pass
 
 
-class Term(aclgenerator.Term):
-  """Representation of an individual term.
+class ServiceMap:
+  """Manages service names across a single policy instance."""
 
-  This is mostly useful for the __str__() method.
+  def __init__(self):
+    self.entries = {}
 
-  Attributes:
-    obj: a policy.Term object
-    term_type: type of filter to generate, e.g. inet or inet6
-    filter_options: list of remaining target options (zones)
-  """
+  def get_service_name(self, term_name, src_ports, ports, protocol, prefix=None):
+    """Returns service name based on the provided ports and protocol."""
+    if (src_ports, ports, protocol) in self.entries:
+      return self.entries[(src_ports, ports, protocol)]["name"]
 
-  ACTIONS = {
-      "accept": "allow",
-      "deny": "deny",
-      "reject": "reject",
-      "count": "count",
-      "log": "log"
-  }
+    if prefix is None:
+      prefix = "service-"
+    service_name = "%s%s-%s" % (prefix, term_name, protocol)
 
-  def __init__(self, term, term_type, zones):
-    self.term = term
-    self.term_type = term_type
-    self.from_zone = zones[1]
-    self.to_zone = zones[3]
-    self.extra_actions = []
+    if len(service_name) > 63:
+      raise PaloAltoFWNameTooLongError(
+          "Service name must be 63 characters max: %s" % service_name)
 
-  def __str__(self):
-    """Render config output from this term object."""
-    # Verify platform specific terms. Skip whole term if platform does not
-    # match.
-    # Nothing here for now
-
-  def _Group(self, group):
-    """If 1 item return it, else return [ item1 item2 ].
-
-    Args:
-      group: a list.  could be a list of strings (protocols) or a list of tuples
-        (ports)
-
-    Returns:
-      rval: a string surrounded by '[' and '];' if len(group) > 1
-            or with just ';' appended if len(group) == 1
-    """
-
-    def _FormattedGroup(el):
-      """Return the actual formatting of an individual element.
-
-      Args:
-        el: either a string (protocol) or a tuple (ports)
-
-      Returns:
-        string: either the lower()'ed string or the ports, hyphenated
-                if they're a range, or by itself if it's not.
-      """
-      if isinstance(el, str):
-        return el.lower()
-      elif isinstance(el, int):
-        return str(el)
-      # type is a tuple below here
-      elif el[0] == el[1]:
-        return "%d" % el[0]
-      else:
-        return "%d-%d" % (el[0], el[1])
-
-    if len(group) > 1:
-      rval = "[ " + " ".join([_FormattedGroup(x) for x in group]) + " ];"
-    else:
-      rval = _FormattedGroup(group[0]) + ";"
-    return rval
-
-
-class Service(object):
-  """Generate PacketFilter policy terms."""
-  service_map = {}
-
-  def __init__(self, ports, service_name,
-               protocol):  # ports is a tuple of ports
-    if (ports, protocol) in self.service_map:
-      raise PaloAltoFWDuplicateServiceError(
-          ("You have a duplicate service. "
-           "A service already exists on port(s): %s") % str(ports))
-
-    final_service_name = "service-" + service_name + "-" + protocol
-
-    for unused_k, v in Service.service_map.items():
-      if v["name"] == final_service_name:
+    for _, service in self.entries.items():
+      if service["name"] == service_name:
         raise PaloAltoFWDuplicateServiceError(
             "You have a duplicate service. A service named %s already exists." %
-            str(final_service_name))
+            service_name)
 
-    if len(final_service_name) > 63:
-      raise PaloAltoFWNameTooLongError(
-          "Service name must be 63 characters max: %s" %
-          str(final_service_name))
-    self.service_map[(ports, protocol)] = {"name": final_service_name}
+    self.entries[(src_ports, ports, protocol)] = {"name": service_name}
+    return service_name
 
 
-class Rule(object):
+class Rule:
   """Extend the Term() class for PaloAlto Firewall Rules."""
 
-  def __init__(self, from_zone, to_zone, terms):
+  def __init__(self, from_zone, to_zone, term, service_map):
     # Palo Alto Firewall rule keys
-    self.options = {}
-    self.options["from_zone"] = [from_zone]
-    self.options["to_zone"] = [to_zone]
+    MAX_ZONE_LENGTH = 31
+
     if not from_zone or not to_zone:
       raise PaloAltoFWOptionError("Source or destination zone is empty.")
+    if len(from_zone) > MAX_ZONE_LENGTH:
+      x = "Source zone must be %d characters max: %s" % (MAX_ZONE_LENGTH,
+                                                         from_zone)
+      raise PaloAltoFWNameTooLongError(x)
+    if len(to_zone) > MAX_ZONE_LENGTH:
+      x = "Destination zone must be %d characters max: %s" % (MAX_ZONE_LENGTH,
+                                                              to_zone)
+      raise PaloAltoFWNameTooLongError(x)
 
-    self.ModifyOptions(terms)
+    self.options = []
 
-  def ModifyOptions(self, terms):
-    """Massage firewall rules into Palo Alto rules format."""
-    term = terms.term
-    self.options["description"] = []
-    self.options["source"] = []
-    self.options["destination"] = []
-    self.options["application"] = []
-    self.options["service"] = []
+    while term is not None:
+      x, term = self.TermToOptions(from_zone, to_zone,
+                                   term, service_map)
+      self.options.append(x)
+
+  @staticmethod
+  def TermToOptions(from_zone, to_zone, term, service_map):
+    """Convert term to Palo Alto security rule options."""
+    options = {}
+    options["from_zone"] = [from_zone]
+    options["to_zone"] = [to_zone]
+    options["description"] = []
+    options["source"] = []
+    options["destination"] = []
+    options["application"] = []
+    options["service"] = []
+    options["logging"] = []
+
+    ACTIONS = {
+      "accept": "allow",
+      "deny": "deny",
+      "reject": "reset-client",
+      "reject-with-tcp-rst": "reset-client",
+    }
+
+    new_term = None
+
+    def pan_ports(ports):
+      x = []
+      for tup in ports:
+        if len(tup) > 1 and tup[0] != tup[1]:
+          x.append(str(tup[0]) + "-" + str(tup[1]))
+        else:
+          x.append(str(tup[0]))
+
+      return tuple(x)
 
     # COMMENT
     if term.comment:
-      self.options["description"] = term.comment
+      options["description"] = term.comment
+
+    # LOGGING
+    if term.logging:
+      for item in term.logging:
+        if item.value in ["disable"]:
+          options["logging"] = ["disable"]
+          break
+        elif item.value in ["log-both"]:
+          options["logging"].append("log-start")
+          options["logging"].append("log-end")
+        elif item.value in ["True", "true", "syslog", "local"]:
+          options["logging"].append("log-end")
+
     # SOURCE-ADDRESS
     if term.source_address:
       saddr_check = set()
@@ -194,9 +174,8 @@ class Rule(object):
         saddr_check.add(saddr.parent_token)
       saddr_check = sorted(saddr_check)
       for addr in saddr_check:
-        self.options["source"].append(str(addr))
-    else:
-      self.options["source"].append("any")
+        options["source"].append(str(addr))
+    # missing source handled during XML document generation
 
     # DESTINATION-ADDRESS
     if term.destination_address:
@@ -205,53 +184,63 @@ class Rule(object):
         daddr_check.add(daddr.parent_token)
       daddr_check = sorted(daddr_check)
       for addr in daddr_check:
-        self.options["destination"].append(str(addr))
-    else:
-      self.options["destination"].append("any")
+        options["destination"].append(str(addr))
+    # missing destination handled during XML document generation
 
+    # ACTION
     if term.action:
-      self.options["action"] = term.action[0]
+      options["action"] = ACTIONS[term.action[0]]
 
     if term.option:
-      self.options["option"] = term.option
+      options["option"] = term.option
 
     if term.pan_application:
       for pan_app in term.pan_application:
-        self.options["application"].append(pan_app)
+        options["application"].append(pan_app)
 
-    if term.destination_port:
-      ports = []
-      for tup in term.destination_port:
-        if len(tup) > 1 and tup[0] != tup[1]:
-          ports.append(str(tup[0]) + "-" + str(tup[1]))
-        else:
-          ports.append(str(tup[0]))
-      ports = tuple(ports)
+    if term.source_port or term.destination_port:
+      src_ports = pan_ports(term.source_port)
+      if term.destination_port:
+        ports = pan_ports(term.destination_port)
+      else:
+        ports = pan_ports([("0", "65535")])
 
       # check to see if this service already exists
       for p in term.protocol:
-        if (ports, p) in Service.service_map:
-          self.options["service"].append(Service.service_map[(ports,
-                                                              p)]["name"])
-        else:
-          # create service
-          unused_new_service = Service(ports, term.name, p)
-          self.options["service"].append(Service.service_map[(ports,
-                                                              p)]["name"])
+        service_name = service_map.get_service_name(term.name,
+                                                    src_ports, ports, p)
+        if service_name not in options["service"]:
+          options["service"].append(service_name)
+
+    elif "tcp" in term.protocol or "udp" in term.protocol:
+      services = {"tcp", "udp"} & set(term.protocol)
+      others = set(term.protocol) - services
+      if others:
+        logging.info(
+          "INFO: Term %s in policy %s>%s contains port-less %s "
+          "with non-port protocol(s). Moving %s to a new term.",
+          term.name, from_zone, to_zone,
+          ', '.join(list(services)), ', '.join(list(others)))
+        new_term = copy.deepcopy(term)
+        new_term.protocol = list(others)
+        term.protocol = list(services)
+        options["application"] = []
+      for p in term.protocol:
+        ports = pan_ports([("0", "65535")])
+        # use prefix "" to avoid service name clash with term named "any"
+        service_name = service_map.get_service_name("any", (), ports, p, "")
+        if service_name not in options["service"]:
+          options["service"].append(service_name)
 
     if term.protocol:
-      # Add application "any" to all terms, unless ICMP/ICMPv6
+      # Add certain protocol names as application in the application list
+      # if missing.
       for proto_name in term.protocol:
-        if proto_name in ["icmp", "icmpv6"]:
-          continue
-        elif proto_name in ["igmp", "sctp"]:
-          if proto_name not in self.options["application"]:
-            self.options["application"].append(proto_name)
-        elif proto_name in ["tcp", "udp"]:
-          if "any" not in self.options["application"]:
-            self.options["application"].append("any")
-        else:
-          pass
+        if (proto_name in ["igmp", "sctp", "gre"] and
+            proto_name not in options["application"]):
+          options["application"].append(proto_name)
+
+    return options, new_term
 
 
 class PaloAltoFW(aclgenerator.ACLGenerator):
@@ -262,8 +251,19 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
   _SUPPORTED_AF = set(("inet", "inet6", "mixed"))
   _AF_MAP = {"inet": (4,), "inet6": (6,), "mixed": (4, 6)}
   _TERM_MAX_LENGTH = 31
-  _SUPPORTED_PROTO_NAMES = ["tcp", "udp", "icmp", "icmpv6", "sctp", "igmp"]
+  _SUPPORTED_PROTO_NAMES = [
+      "tcp",
+      "udp",
+      "icmp",
+      "icmpv6",
+      "sctp",
+      "igmp",
+      "gre",
+  ]
   _MAX_RULE_DESCRIPTION_LENGTH = 1024
+  _MAX_TAG_COMMENTS_LENGTH = 1023
+  _TAG_NAME_FORMAT = "{from_zone}_{to_zone}_policy-comment-{num}"
+  _MAX_RULE_SRC_DST_MEMBERS = 65535
 
   INDENT = "  "
 
@@ -278,7 +278,9 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
     self.from_zone = ""
     self.to_zone = ""
     self.policy_name = ""
-    super(PaloAltoFW, self).__init__(pol, exp_info)
+    self.config = None
+    self.service_map = ServiceMap()
+    super().__init__(pol, exp_info)
 
   def _BuildTokens(self):
     """Build supported tokens for platform.
@@ -286,8 +288,7 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
     Returns:
       tuple containing both supported tokens and sub tokens
     """
-    supported_tokens, supported_sub_tokens = super(PaloAltoFW,
-                                                   self)._BuildTokens()
+    supported_tokens, supported_sub_tokens = super()._BuildTokens()
 
     supported_tokens = {
         "action",
@@ -302,6 +303,7 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
         "option",
         "owner",
         "platform",
+        "platform_exclude",
         "protocol",
         "source_address",
         "source_address_exclude",
@@ -313,7 +315,7 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
     }
 
     supported_sub_tokens.update({
-        "action": {"accept", "deny", "reject", "count", "log"},
+        "action": {"accept", "deny", "reject", "reject-with-tcp-rst"},
         "option": {"established", "tcp-established"},
     })
     return supported_tokens, supported_sub_tokens
@@ -339,6 +341,8 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
     """
     current_date = datetime.date.today()
     exp_info_date = current_date + datetime.timedelta(weeks=exp_info)
+    first_addr_obj = None
+
     for header, terms in pol.filters:
       if self._PLATFORM not in header.platforms:
         continue
@@ -366,8 +370,22 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
 
       if filter_type not in self._SUPPORTED_AF:
         raise UnsupportedHeaderError(
-            "Palo Alto Firewall Generator currently does not support"
-            " %s as a header option" % (filter_type))
+            'Palo Alto Firewall Generator invalid address family option: "%s"'
+            '; expect {%s}' % (filter_type, '|'.join(self._SUPPORTED_AF)))
+
+      valid_addr_obj = ["addr-obj", "no-addr-obj"]
+      if len(filter_options) > 5 and filter_options[5] not in valid_addr_obj:
+        raise UnsupportedHeaderError(
+            'Palo Alto Firewall Generator invalid address objects option: "%s"'
+            '; expect {%s}' % (filter_options[5], '|'.join(valid_addr_obj)))
+      no_addr_obj = True if (len(filter_options) > 5 and
+                             filter_options[5] == "no-addr-obj") else False
+      if first_addr_obj is None:
+        first_addr_obj = no_addr_obj
+      if first_addr_obj != no_addr_obj:
+        raise UnsupportedHeaderError(
+            "Cannot mix addr-obj and no-addr-obj header option in "
+            "a single policy file")
 
       term_dup_check = set()
       new_terms = []
@@ -391,12 +409,30 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
               "term and will not be rendered.", term.name, self.from_zone,
               self.to_zone)
           continue
+
+        # Verify platform specific terms. Skip whole term if platform does not
+        # match.
+        if term.platform:
+          if self._PLATFORM not in term.platform:
+            continue
+        if term.platform_exclude:
+          if self._PLATFORM in term.platform_exclude:
+            continue
+
         term.name = self.FixTermLength(term.name)
         if term.name in term_dup_check:
           raise PaloAltoFWDuplicateTermError("You have a duplicate term: %s" %
                                              term.name)
         term_dup_check.add(term.name)
 
+        services = {"tcp", "udp"} & set(term.protocol)
+        others = set(term.protocol) - services
+        if others and term.pan_application:
+          raise UnsupportedFilterError(
+            "Term %s contains non tcp, udp protocols with pan-application: %s: %s"
+            "\npan-application can only be used with protocols tcp, udp" %
+            (term.name, ', '.join(term.pan_application), ', '.join(term.protocol)))
+
         if term.expiration:
           if term.expiration <= exp_info_date:
             logging.info(
@@ -539,6 +575,9 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
           if ("icmp" not in term.protocol and "icmpv6" not in term.protocol):
             # the protocol is not ICMP or ICMPv6
             break
+          if icmp_version not in term.protocol:
+            # skip if this icmp_version isn't in the term protocol.
+            continue
           if icmp_version == "icmp" and "ip4-ip4" not in flows:
             # skip if there is no ip4 to ipv4 communication
             continue
@@ -604,19 +643,32 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
           raise PaloAltoFWUnsupportedProtocolError(
               "protocol %s is not supported" % proto_name)
 
-        # Create Term object with the term, address family, and header
-        # parameters, e.g. to/from zone, and add it to a list of
-        # terms that would form a rule.
-        new_term = Term(term, filter_type, filter_options)
-        new_terms.append(new_term)
+        if term.icmp_type:
+          if set(term.protocol) == {'icmp', 'icmpv6'}:
+            raise UnsupportedFilterError('%s %s' % (
+                'icmp-type specified for both icmp and icmpv6 protocols'
+                ' in a single term:', term.name))
+          if term.protocol != ['icmp'] and term.protocol != ['icmpv6']:
+            raise UnsupportedFilterError('%s %s' % (
+                'icmp-type specified for non-icmp protocols in term:',
+                term.name))
+
+        new_terms.append(term)
 
       # Create a ruleset. It contains the rules for the terms defined under
       # a single header on a particular platform.
       ruleset = {}
 
       for term in new_terms:
-        current_rule = Rule(self.from_zone, self.to_zone, term)
-        ruleset[term.term.name] = current_rule.options
+        current_rule = Rule(self.from_zone, self.to_zone, term,
+                            self.service_map)
+        if len(current_rule.options) > 1:
+          for i, v in enumerate(current_rule.options):
+            name = "%s-%d" % (term.name, i+1)
+            name = self.FixTermLength(name)
+            ruleset[name] = v
+        else:
+          ruleset[term.name] = current_rule.options[0]
 
       self.pafw_policies.append((header, ruleset, filter_options))
 
@@ -677,14 +729,52 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
 
   def __str__(self):
     """Render the output of the PaloAltoFirewall policy into config."""
-    initial = []
+
+    # IPv4 addresses are normalized into the policy as IPv6 addresses
+    # using ::<ipv4-address>.  The 0.0.0.0-255.255.255.255 range is
+    # equivalent to ::0/96 which will only match IPv4 addresses; when
+    # negated it will match only IPv6 addresses.
+    # Used for address families inet and inet6 when source and
+    # destination address are not specified (any any).
+    ANY_IPV4_RANGE = "0.0.0.0-255.255.255.255"
+    add_any_ipv4 = False
+
+    address_book_names_dict = {}
+    address_book_groups_dict = {}
+    for zone in self.addressbook:
+      # building individual addresses dictionary
+      groups = sorted(self.addressbook[zone])
+      for group in groups:
+        for address, name in self.addressbook[zone][group]:
+          if name in address_book_names_dict:
+            if address_book_names_dict[name].supernet_of(address):
+              continue
+          address_book_names_dict[name] = address
+
+        # building individual address-group dictionary
+        for nested_group in groups:
+          group_names = []
+          for address, name in self.addressbook[zone][nested_group]:
+            group_names.append(name)
+          address_book_groups_dict[nested_group] = group_names
+
+      # sort address books and address sets
+      address_book_groups_dict = collections.OrderedDict(
+          sorted(address_book_groups_dict.items()))
+
+    address_book_keys = sorted(
+        list(address_book_names_dict.keys()), key=self._SortAddressBookNumCheck)
+
     # INITAL CONFIG
-    initial.append('<?xml version="1.0"?>')
-    initial.append('<config version="7.0.0" urldb="paloaltonetworks">')
-    initial.append(self.INDENT * 1 + "<devices>")
-    initial.append(self.INDENT * 2 + '<entry name="localhost.localdomain">')
-    initial.append(self.INDENT * 3 + "<vsys>")
-    initial.append(self.INDENT * 4 + '<entry name="vsys1">')
+    config = etree.Element("config", {
+        "version": "8.1.0",
+        "urldb": "paloaltonetworks"
+    })
+    devices = etree.SubElement(config, "devices")
+    device_entry = etree.SubElement(devices, "entry",
+                                    {"name": "localhost.localdomain"})
+    vsys = etree.SubElement(device_entry, "vsys")
+    vsys_entry = etree.SubElement(vsys, "entry", {"name": "vsys1"})
 
     # APPLICATION
     app_entries = etree.Element("application")
@@ -710,199 +800,245 @@ class PaloAltoFW(aclgenerator.ACLGenerator):
               etree.SubElement(icmp_type_props, "type").text = app[k][prop]
             else:
               pass
-
-    lines = minidom.parseString(
-        etree.tostring(app_entries)).toprettyxml(indent="  ")
-    # TODO(greenpau): currently, the generator relies on xml package only
-    # in the application section. The xml package adds xml header to its
-    # output. At the moment, it needs to be stripped, because the header
-    # already exists.
-    for line in lines.split("\n")[1:]:
-      initial.append(self.INDENT * 5 + str(line))
+    vsys_entry.append(app_entries)
 
     # APPLICATION GROUPS
-    app_group_entries = []
-    if self.application_groups:
-      app_group_entries.append(self.INDENT * 5 + "<application-group>")
-      for app_group in self.application_groups:
-        app_group_entries.append(self.INDENT * 6 + ("%s" % app_group))
-      app_group_entries.append(self.INDENT * 5 + "</application-group>")
-    else:
-      app_group_entries.append(self.INDENT * 5 + "<application-group/>")
-
-    # ADDRESS
-    address_entries = []
-    address_entries.append(self.INDENT * 5 + "<!-- Addresses -->")
-    address_entries.append(self.INDENT * 5 + "<address>")
-
-    address_book_names_dict = {}
-    address_book_groups_dict = {}
-    for zone in self.addressbook:
-      # building individual addresses dictionary
-      groups = sorted(self.addressbook[zone])
-      for group in groups:
-        for address, name in self.addressbook[zone][group]:
-          if name in address_book_names_dict:
-            if address_book_names_dict[name].supernet_of(address):
-              continue
-          address_book_names_dict[name] = address
-
-        # building individual address-group dictionary
-        for nested_group in groups:
-          group_names = []
-          for address, name in self.addressbook[zone][nested_group]:
-            group_names.append(name)
-          address_book_groups_dict[nested_group] = group_names
-
-      # sort address books and address sets
-      address_book_groups_dict = collections.OrderedDict(
-          sorted(address_book_groups_dict.items()))
-    address_book_keys = sorted(
-        list(address_book_names_dict.keys()), key=self._SortAddressBookNumCheck)
-
-    for name in address_book_keys:
-      address_entries.append(self.INDENT * 6 + '<entry name="' + name + '">')
-      address_entries.append(self.INDENT * 7 + "<description>" + name +
-                             "</description>")
-      address_entries.append(self.INDENT * 7 + "<ip-netmask>" +
-                             str(address_book_names_dict[name]) +
-                             "</ip-netmask>")
-      address_entries.append(self.INDENT * 6 + "</entry>")
-
-    address_entries.append(self.INDENT * 5 + "</address>")
-
-    address_group_entries = []
-    address_group_entries.append(self.INDENT * 5 + "<!-- Address groups-->")
-    address_group_entries.append(self.INDENT * 5 + "<address-group>")
-
-    for group, address_list in address_book_groups_dict.items():
-      address_group_entries.append(self.INDENT * 6 + '<entry name="' + group +
-                                   '">')
-      address_group_entries.append(self.INDENT * 7 + "<static>")
-      for name in address_list:
-        address_group_entries.append(self.INDENT * 8 + "<member>" + name +
-                                     "</member>")
-      address_group_entries.append(self.INDENT * 7 + "</static>")
-      address_group_entries.append(self.INDENT * 7 + "</entry>")
-    address_group_entries.append(self.INDENT * 5 + "</address-group>")
+    etree.SubElement(vsys_entry, "application-group")
 
     # SERVICES
-    service = []
-    service.append(self.INDENT * 5 + "<!-- Services -->")
-
-    if Service.service_map.items():
-      service.append(self.INDENT * 5 + "<service>")
-      for k, v in Service.service_map.items():
-        service.append(self.INDENT * 6 + '<entry name="' + v["name"] + '">')
-        service.append(self.INDENT * 7 + "<protocol>")
-        service.append(self.INDENT * 8 + "<" + k[1] + ">")
+    vsys_entry.append(etree.Comment(" Services "))
+    service = etree.SubElement(vsys_entry, "service")
+    for k, v in self.service_map.entries.items():
+      entry = etree.SubElement(service, "entry", {"name": v["name"]})
+      proto0 = etree.SubElement(entry, "protocol")
+      proto = etree.SubElement(proto0, k[2])
+      # destination port
+      port = etree.SubElement(proto, "port")
+      tup = str(k[1])[1:-1]
+      if tup[-1] == ",":
+        tup = tup[:-1]
+      port.text = tup.replace("'", "").replace(", ", ",")
+      # source port
+      if len(k[0]):
+        sport = etree.SubElement(proto, "source-port")
         tup = str(k[0])[1:-1]
         if tup[-1] == ",":
           tup = tup[:-1]
-        service.append(self.INDENT * 9 + "<port>" +
-                       tup.replace("'", "").replace(", ", ",") + "</port>")
-        service.append(self.INDENT * 8 + "</" + k[1] + ">")
-        service.append(self.INDENT * 7 + "</protocol>")
-        service.append(self.INDENT * 6 + "</entry>")
-      service.append(self.INDENT * 5 + "</service>")
-    else:
-      service.append(self.INDENT * 5 + "<service/>")
+        sport.text = tup.replace("'", "").replace(", ", ",")
 
     # RULES
-    rules = []
-    rules.append(self.INDENT * 5 + "<!-- Rules -->")
+    vsys_entry.append(etree.Comment(" Rules "))
+    rulebase = etree.SubElement(vsys_entry, "rulebase")
+    security = etree.SubElement(rulebase, "security")
+    rules = etree.SubElement(security, "rules")
+    tag = etree.Element("tag")
 
-    rules.append(self.INDENT * 5 + "<rulebase>")
-    rules.append(self.INDENT * 6 + "<security>")
-    rules.append(self.INDENT * 7 + "<rules>")
+    tag_num = 0
 
     # pytype: disable=key-error
     # pylint: disable=unused-variable
     for (header, pa_rules, filter_options) in self.pafw_policies:
+      tag_name = None
+      if header.comment:
+        comment = " ".join(header.comment).strip()
+        if comment:
+          tag_num += 1
+          # max tag len 127, max zone len 31
+          tag_name = self._TAG_NAME_FORMAT.format(
+              from_zone=filter_options[1],
+              to_zone=filter_options[3],
+              num=tag_num)
+          tag_entry = etree.SubElement(tag, "entry", {"name": tag_name})
+          comments = etree.SubElement(tag_entry, "comments")
+          if len(comment) > self._MAX_TAG_COMMENTS_LENGTH:
+            logging.warning(
+                "WARNING: tag %s comments exceeds maximum "
+                "length %d, truncated.", tag_name,
+                self._MAX_TAG_COMMENTS_LENGTH)
+          comments.text = comment[:self._MAX_TAG_COMMENTS_LENGTH]
+
+      no_addr_obj = True if (len(filter_options) > 5 and
+                             filter_options[5] == "no-addr-obj") else False
+
       for name, options in pa_rules.items():
-        rules.append(self.INDENT * 8 + '<entry name="' + name + '">')
+        entry = etree.SubElement(rules, "entry", {"name": name})
         if options["description"]:
-          descr = " ".join(options["description"])
-          if len(descr) > self._MAX_RULE_DESCRIPTION_LENGTH:
-            descr = descr[:self._MAX_RULE_DESCRIPTION_LENGTH]
-          rules.append(self.INDENT * 9 + "<description>" +
-                       descr + "</description>")
+          descr = etree.SubElement(entry, "description")
+          x = " ".join(options["description"])
+          if len(x) > self._MAX_RULE_DESCRIPTION_LENGTH:
+            logging.warning(
+                "WARNING: rule %s description exceeds maximum "
+                "length %d, truncated.", name,
+                self._MAX_RULE_DESCRIPTION_LENGTH)
+          descr.text = x[:self._MAX_RULE_DESCRIPTION_LENGTH]
 
-        rules.append(self.INDENT * 9 + "<to>")
-        for tz in options["to_zone"]:
-          rules.append(self.INDENT * 10 + "<member>" + tz + "</member>")
-        rules.append(self.INDENT * 9 + "</to>")
+        to = etree.SubElement(entry, "to")
+        for x in options["to_zone"]:
+          member = etree.SubElement(to, "member")
+          member.text = x
 
-        rules.append(self.INDENT * 9 + "<from>")
-        for fz in options["from_zone"]:
-          rules.append(self.INDENT * 10 + "<member>" + fz + "</member>")
-        rules.append(self.INDENT * 9 + "</from>")
+        from_ = etree.SubElement(entry, "from")
+        for x in options["from_zone"]:
+          member = etree.SubElement(from_, "member")
+          member.text = x
 
-        rules.append(self.INDENT * 9 + "<source>")
+        af = filter_options[4] if len(filter_options) > 4 else "inet"
+
+        max_src_dst = 0
+        source = etree.SubElement(entry, "source")
         if not options["source"]:
-          rules.append(self.INDENT * 10 + "<member>any</member>")
+          member = etree.SubElement(source, "member")
+          if not options["destination"] and af != "mixed":
+            # only inet and inet6 use the any-ipv4 object
+            member.text = "any-ipv4"
+            add_any_ipv4 = True
+          else:
+            member.text = "any"
         else:
-          for s in options["source"]:
-            rules.append(self.INDENT * 10 + "<member>" + s + "</member>")
-        rules.append(self.INDENT * 9 + "</source>")
+          for x in options["source"]:
+            if no_addr_obj:
+              for group in address_book_groups_dict[x]:
+                member = etree.SubElement(source, "member")
+                member.text = str(address_book_names_dict[group])
+                max_src_dst += 1
+            else:
+              member = etree.SubElement(source, "member")
+              member.text = x
+              max_src_dst += 1
 
-        rules.append(self.INDENT * 9 + "<destination>")
+        if max_src_dst > self._MAX_RULE_SRC_DST_MEMBERS:
+          raise UnsupportedFilterError(
+            "term %s source members exceeds maximum of %d: %d" %
+            (name, self._MAX_RULE_SRC_DST_MEMBERS, max_src_dst))
+
+        max_src_dst = 0
+        dest = etree.SubElement(entry, "destination")
         if not options["destination"]:
-          rules.append(self.INDENT * 10 + "<member>any</member>")
+          member = etree.SubElement(dest, "member")
+          if options["source"]:
+            member.text = "any"
+          else:
+            if af != "mixed":
+              # only inet and inet6 use the any-ipv4 object
+              member.text = "any-ipv4"
+              if af == "inet6":
+                for x in ["negate-source", "negate-destination"]:
+                  negate = etree.SubElement(entry, x)
+                  negate.text = "yes"
+            else:
+              member.text = "any"
         else:
-          for d in options["destination"]:
-            rules.append(self.INDENT * 10 + "<member>" + d + "</member>")
-        rules.append(self.INDENT * 9 + "</destination>")
+          for x in options["destination"]:
+            if no_addr_obj:
+              for group in address_book_groups_dict[x]:
+                member = etree.SubElement(dest, "member")
+                member.text = str(address_book_names_dict[group])
+                max_src_dst += 1
+            else:
+              member = etree.SubElement(dest, "member")
+              member.text = x
+              max_src_dst += 1
+
+        if max_src_dst > self._MAX_RULE_SRC_DST_MEMBERS:
+          raise UnsupportedFilterError(
+            "term %s destination members exceeds maximum of %d: %d" %
+            (name, self._MAX_RULE_SRC_DST_MEMBERS, max_src_dst))
 
         # service section of a policy rule.
-        rules.append(self.INDENT * 9 + "<service>")
+        service = etree.SubElement(entry, "service")
         if not options["service"] and not options["application"]:
-          rules.append(self.INDENT * 10 + "<member>any</member>")
+          member = etree.SubElement(service, "member")
+          member.text = "any"
         elif not options["service"] and options["application"]:
           # Adds custom applications.
-          rules.append(self.INDENT * 10 +
-                       "<member>application-default</member>")
+          member = etree.SubElement(service, "member")
+          member.text = "application-default"
         else:
           # Adds services.
-          for s in options["service"]:
-            rules.append(self.INDENT * 10 + "<member>" + s + "</member>")
-        rules.append(self.INDENT * 9 + "</service>")
+          for x in options["service"]:
+            member = etree.SubElement(service, "member")
+            member.text = x
 
         # ACTION
-        rules.append(self.INDENT * 9 + "<action>" +
-                     Term.ACTIONS.get(str(options["action"])) + "</action>")
+        action = etree.SubElement(entry, "action")
+        action.text = options["action"]
 
         # check whether the rule is interzone
         if list(set(options["from_zone"]).difference(options["to_zone"])):
-          rules.append(self.INDENT * 9 + "<rule-type>interzone</rule-type>")
+          type_ = etree.SubElement(entry, "rule-type")
+          type_.text = "interzone"
         elif not options["from_zone"] and not options["to_zone"]:
-          rules.append(self.INDENT * 9 + "<rule-type>interzone</rule-type>")
+          type_ = etree.SubElement(entry, "rule-type")
+          type_.text = "interzone"
 
         # APPLICATION
-        rules.append(self.INDENT * 9 + "<application>")
+        app = etree.SubElement(entry, "application")
         if not options["application"]:
-          rules.append(self.INDENT * 10 + "<member>any</member>")
+          member = etree.SubElement(app, "member")
+          member.text = "any"
         else:
-          for a in options["application"]:
-            rules.append(self.INDENT * 10 + "<member>" + a + "</member>")
-        rules.append(self.INDENT * 9 + "</application>")
+          for x in options["application"]:
+            member = etree.SubElement(app, "member")
+            member.text = x
 
-        rules.append(self.INDENT * 8 + "</entry>")
+        if tag_name is not None:
+          rules_tag = etree.SubElement(entry, "tag")
+          member = etree.SubElement(rules_tag, "member")
+          member.text = tag_name
+
+        # LOGGING
+        if options["logging"]:
+          if "disable" in options["logging"]:
+            log = etree.SubElement(entry, "log-start")
+            log.text = "no"
+            log = etree.SubElement(entry, "log-end")
+            log.text = "no"
+          if "log-start" in options["logging"]:
+            log = etree.SubElement(entry, "log-start")
+            log.text = "yes"
+          if "log-end" in options["logging"]:
+            log = etree.SubElement(entry, "log-end")
+            log.text = "yes"
 
     # pytype: enable=key-error
-    rules.append(self.INDENT * 7 + "</rules>")
-    rules.append(self.INDENT * 6 + "</security>")
 
-    rules.append(self.INDENT * 5 + "</rulebase>")
+    if no_addr_obj:
+      address_book_groups_dict = {}
+      address_book_keys = {}
 
-    end = []
-    end.append(self.INDENT * 4 + "</entry>")
-    end.append(self.INDENT * 3 + "</vsys>")
-    end.append(self.INDENT * 2 + "</entry>")
-    end.append(self.INDENT * 1 + "</devices>")
-    end.append("</config>\n")
+    # ADDRESS
+    vsys_entry.append(etree.Comment(" Address Groups "))
+    addr_group = etree.SubElement(vsys_entry, "address-group")
 
-    return ("\n".join(initial) + "\n\n" + "\n".join(app_group_entries) +
-            "\n\n" + "\n".join(service) + "\n\n" + "\n".join(rules) + "\n\n" +
-            "\n".join(address_group_entries) + "\n\n" +
-            "\n".join(address_entries) + "\n" + "\n".join(end))
+    for group, address_list in address_book_groups_dict.items():
+      entry = etree.SubElement(addr_group, "entry", {"name": group})
+      static = etree.SubElement(entry, "static")
+      for name in address_list:
+        member = etree.SubElement(static, "member")
+        member.text = name
+
+    vsys_entry.append(etree.Comment(" Addresses "))
+    addr = etree.SubElement(vsys_entry, "address")
+
+    for name in address_book_keys:
+      entry = etree.SubElement(addr, "entry", {"name": name})
+      desc = etree.SubElement(entry, "description")
+      desc.text = name
+      ip = etree.SubElement(entry, "ip-netmask")
+      ip.text = str(address_book_names_dict[name])
+
+    if add_any_ipv4:
+      entry = etree.SubElement(addr, "entry", {"name": "any-ipv4"})
+      desc = etree.SubElement(entry, "description")
+      desc.text = ("Object to match all IPv4 addresses; "
+                   "negate to match all IPv6 addresses.")
+      range = etree.SubElement(entry, "ip-range")
+      range.text = ANY_IPV4_RANGE
+
+    vsys_entry.append(tag)
+
+    self.config = config
+    document = etree.tostring(config, encoding="UTF-8")
+    dom = minidom.parseString(document.decode("UTF-8"))
+
+    return dom.toprettyxml(indent=self.INDENT)
