@@ -16,11 +16,6 @@
 """Parses the generic policy files and return a policy object for acl rendering.
 """
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
 import datetime
 import os
 import sys
@@ -30,8 +25,6 @@ from capirca.lib import nacaddr
 from capirca.lib import naming
 from ply import lex
 from ply import yacc
-from six.moves import map
-from six.moves import range
 
 
 DEFINITIONS = None
@@ -166,7 +159,7 @@ def TranslatePorts(ports, protocols, term_name):
 
 
 # classes for storing the object types in the policy files.
-class Policy(object):
+class Policy:
   """The policy object contains everything found in a given policy file."""
 
   def __init__(self, header, terms):
@@ -302,7 +295,7 @@ class Policy(object):
     return self.__str__()
 
 
-class Term(object):
+class Term:
   """The Term object is used to store each of the terms.
 
   Args:
@@ -313,6 +306,7 @@ class Term(object):
       VarType.(S|D)?ADDRESS's
     address_exclude/source_address_exclude/destination_address_exclude: list of
       VarType.(S|D)?ADDEXCLUDE's
+    restrict-address-family: VarType.RESTRICT_ADDRESS_FAMILY
     port/source_port/destination_port: list of VarType.(S|D)?PORT's
     options: list of VarType.OPTION's.
     protocol: list of VarType.PROTOCOL's.
@@ -324,6 +318,7 @@ class Term(object):
     dscp-except: VarType.DSCP_EXCEPT
     comments: VarType.COMMENT
     encapsulate: VarType.ENCAPSULATE
+    filter-term: VarType.FILTER_TERM
     flexible-match-range: VarType.FLEXIBLE_MATCH_RANGE
     forwarding-class: VarType.FORWARDING_CLASS
     forwarding-class-except: VarType.FORWARDING_CLASS_EXCEPT
@@ -409,6 +404,7 @@ class Term(object):
     self.action = []
     self.address = []
     self.address_exclude = []
+    self.restrict_address_family = None
     self.comment = []
     self.counter = None
     self.expiration = None
@@ -416,6 +412,7 @@ class Term(object):
     self.destination_address_exclude = []
     self.destination_port = []
     self.destination_prefix = []
+    self.filter_term = None
     self.forwarding_class = []
     self.forwarding_class_except = []
     self.logging = []
@@ -714,6 +711,8 @@ class Term(object):
     if self.destination_prefix_except:
       ret_str.append('  destination_prefix_except: %s' %
                      self.destination_prefix_except)
+    if self.filter_term:
+      ret_str.append('  filter_term: %s' % self.filter_term)
     if self.forwarding_class:
       ret_str.append('  forwarding_class: %s' % self.forwarding_class)
     if self.forwarding_class_except:
@@ -901,6 +900,10 @@ class Term(object):
     if self.precedence != other.precedence:
       return False
 
+    # filter
+    if self.filter_term != other.filter_term:
+      return False
+
     # forwarding-class
     if sorted(self.forwarding_class) != sorted(other.forwarding_class):
       return False
@@ -1121,7 +1124,9 @@ class Term(object):
                   type(x), x.value))
     else:
       # stupid no switch statement in python
-      if obj.var_type is VarType.COMMENT:
+      if obj.var_type is VarType.RESTRICT_ADDRESS_FAMILY:
+        self.restrict_address_family = obj.value
+      elif obj.var_type is VarType.COMMENT:
         self.comment.append(str(obj))
       elif obj.var_type is VarType.OWNER:
         self.owner = obj.value
@@ -1196,6 +1201,8 @@ class Term(object):
         self.target_resources.append(obj.value)
       elif obj.var_type is VarType.TARGET_SERVICE_ACCOUNTS:
         self.target_service_accounts.append(obj.value)
+      elif obj.var_type is VarType.FILTER_TERM:
+        self.filter_term = obj.value
       else:
         raise TermObjectTypeError(
             '%s isn\'t a type I know how to deal with' % (type(obj)))
@@ -1229,8 +1236,11 @@ class Term(object):
         raise ParseError(
             'term "%s" has both verbatim and non-verbatim tokens.' % self.name)
     else:
-      if not self.action and not self.routing_instance and not self.next_ip and not self.encapsulate:
+      if (not self.action and not self.routing_instance and not self.next_ip and
+          not self.encapsulate and not self.filter_term):
         raise TermNoActionError('no action specified for term %s' % self.name)
+      if self.filter_term and self.action:
+        raise InvalidTermActionError('term "%s" has both filter and action tokens.' % self.name)
       # have we specified a port with a protocol that doesn't support ports?
       if self.source_port or self.destination_port or self.port:
         if not any(proto in self.protocol for proto in ['tcp', 'udp', 'sctp']):
@@ -1436,7 +1446,7 @@ class Term(object):
     return True
 
 
-class VarType(object):
+class VarType:
   """Generic object meant to store lots of basic policy types."""
 
   COMMENT = 0
@@ -1500,6 +1510,9 @@ class VarType(object):
   TARGET_RESOURCES = 59
   TARGET_SERVICE_ACCOUNTS = 60
   ENCAPSULATE = 61
+  FILTER_TERM = 62
+  RESTRICT_ADDRESS_FAMILY = 63
+
 
   def __init__(self, var_type, value):
     self.var_type = var_type
@@ -1524,7 +1537,7 @@ class VarType(object):
     return id(self)
 
 
-class Header(object):
+class Header:
   """The header of the policy file contains the targets and a global comment."""
 
   def __init__(self):
@@ -1632,7 +1645,7 @@ class Header(object):
 # b/c we're almost certainly going to have to do something more exotic with
 # it shortly to account for various rendering options like default iptables
 # policies or output file names, etc. etc.
-class Target(object):
+class Target:
   """The type of acl to be rendered from this policy file."""
 
   def __init__(self, target):
@@ -1657,6 +1670,7 @@ tokens = (
     'ACTION',
     'ADDR',
     'ADDREXCLUDE',
+    'RESTRICT_ADDRESS_FAMILY',
     'COMMENT',
     'COUNTER',
     'DADDR',
@@ -1676,6 +1690,7 @@ tokens = (
     'ESCAPEDSTRING',
     'ETHER_TYPE',
     'EXPIRATION',
+    'FILTER_TERM',
     'FLEXIBLE_MATCH_RANGE',
     'FORWARDING_CLASS',
     'FORWARDING_CLASS_EXCEPT',
@@ -1742,6 +1757,7 @@ reserved = {
     'action': 'ACTION',
     'address': 'ADDR',
     'address-exclude': 'ADDREXCLUDE',
+    'restrict-address-family': 'RESTRICT_ADDRESS_FAMILY',
     'comment': 'COMMENT',
     'counter': 'COUNTER',
     'destination-address': 'DADDR',
@@ -1757,6 +1773,7 @@ reserved = {
     'encapsulate': 'ENCAPSULATE',
     'ether-type': 'ETHER_TYPE',
     'expiration': 'EXPIRATION',
+    'filter-term': 'FILTER_TERM',
     'flexible-match-range': 'FLEXIBLE_MATCH_RANGE',
     'forwarding-class': 'FORWARDING_CLASS',
     'forwarding-class-except': 'FORWARDING_CLASS_EXCEPT',
@@ -1928,6 +1945,7 @@ def p_terms(p):
 def p_term_spec(p):
   """ term_spec : term_spec action_spec
                 | term_spec addr_spec
+                | term_spec restrict_address_family_spec
                 | term_spec comment_spec
                 | term_spec counter_spec
                 | term_spec traffic_class_count_spec
@@ -1938,6 +1956,7 @@ def p_term_spec(p):
                 | term_spec ether_type_spec
                 | term_spec exclude_spec
                 | term_spec expiration_spec
+                | term_spec filter_term_spec
                 | term_spec flexible_match_range_spec
                 | term_spec forwarding_class_spec
                 | term_spec forwarding_class_except_spec
@@ -1980,6 +1999,9 @@ def p_term_spec(p):
     else:
       p[0] = Term(p[2])
 
+def p_restrict_address_family_spec(p):
+  """ restrict_address_family_spec : RESTRICT_ADDRESS_FAMILY ':' ':' STRING """
+  p[0] = VarType(VarType.RESTRICT_ADDRESS_FAMILY, p[4])
 
 def p_routinginstance_spec(p):
   """ routinginstance_spec : ROUTING_INSTANCE ':' ':' STRING """
@@ -2376,6 +2398,9 @@ def p_ttl_spec(p):
   """ ttl_spec : TTL ':' ':' INTEGER """
   p[0] = VarType(VarType.TTL, p[4])
 
+def p_filter_term_spec(p):
+  """ filter_term_spec : FILTER_TERM ':' ':' STRING """
+  p[0] = VarType(VarType.FILTER_TERM, p[4])
 
 def p_one_or_more_strings(p):
   """ one_or_more_strings : one_or_more_strings STRING
